//SETTINGS
//Sonic's drop dash
local DROP_LV1TIME = 10
local DROP_LV2TIME = 45
local DROP_LV3TIME = 85
local DROP_LV1SPEED = 40*FRACUNIT
local DROP_LV2SPEED = 60*FRACUNIT
local DROP_LV3SPEED = 80*FRACUNIT

//Sonic's double jump
local DJ_SPEED = 8*FRACUNIT
local DJ_MOMZ = (19/2)*FRACUNIT

//Tails's flight
local FLY_AMOUNT = 3
local FLY_MOMZ = (53/20)*FRACUNIT
local FLY_MOMZTICS = 7
local FLY_MOMZTICS_SWIM = 4
local FLY_SPAMTICS = 10
local FLY_HOVER = -(7/2) * FRACUNIT
local FLY_HOVER_TIRED = -(11/2) * FRACUNIT
local FLY_TIREDTIME = 55
local FLY_MAXTIME = 85
local FLY_MAXTIMEWATER = 25
local FLY_MAXMOMZ = 10 * FRACUNIT

//Knuckles's glide
local GLIDE_MAXMOMZ = 10 * FRACUNIT

//Misc
local UNCURL_LOCKTIME = 8
local BOSS_MOMENTUM_PERCENT = 75

local function ctrlStyle(p)
    -- Returns 3 if "Simple" control scheme is enabled.
    -- Returns 2 if directionchar is set to movement
    -- Returns 1 if directionchar is set to camera, but Simple control scheme is enabled, AKA Analog Legacy
    -- Returns 0 if directionchar is set to camera, AKA Legacy
    return ((p.pflags & PF_ANALOGMODE) and 1 or 0) | ((p.pflags & PF_DIRECTIONCHAR) and 2 or 0)
end

COM_AddCommand("freeaim", function(player, value)
	if not(player and player.valid and player.mo)
		CONS_Printf(player,"This command can only be used inside a level.")
		return
	end
	if value == "0" or value == "no" or value == "off"
		player.freeaim = 0
		CONS_Printf(player,"Free ability aim disabled.")
	elseif value == "1" or value == "yes" or value == "on"
		player.freeaim = 1
		CONS_Printf(player,"Free ability aim enabled.")
	else
		if player.freeaim == 1
			CONS_Printf(player,"freeaim is on. Default is on.")
		else
			CONS_Printf(player,"freeaim is off. Default is on.")
		end
	end
end,0)

COM_AddCommand("uncurltoggle", function(player, value)
	if not(player and player.valid and player.mo)
		CONS_Printf(player,"This command can only be used inside a level.")
		return
	end
	if value == "0" or value == "no" or value == "off"
		player.uncurltoggle = 0
		CONS_Printf(player,"Uncurl toggle disabled.")
	elseif value == "1" or value == "yes" or value == "on"
		player.uncurltoggle = 1
		CONS_Printf(player,"Uncurl toggle enabled.")
	else
		if player.uncurltoggle == 1
			CONS_Printf(player,"uncurltoggle is on. Default is on.")
		else
			CONS_Printf(player,"uncurltoggle is off. Default is on.")
		end
	end
end,0)

//Restore all abilities
local function resetAbilities(player)
	if player and player.mo and player.mo.valid
		local mo = player.mo
		if mo.dropcharge == nil mo.dropcharge = 0 end
		mo.flytimer = 0
		mo.flyamount = FLY_AMOUNT
		mo.djready = true
		player.pflags = $ & ~PF_THOKKED
	end
end

//Sonic double jump activated
local function sonicDoubleJump(player)
	local mo = player.mo
	
	if not mo.djready return end
	S_StartSound(mo, sfx_dbjump)
	
	local pbtn = player.prevbuttons
	local btn = player.cmd.buttons
	local speed = FixedHypot(mo.momx, mo.momy)
	local momangle = R_PointToAngle2(0,0,mo.momx,mo.momy)
	local djspeed = DJ_SPEED
	local djmomz = DJ_MOMZ
	local dustamt = 4
	local abilityaim = mo.angle + R_PointToAngle2(0,0,player.cmd.forwardmove * FRACUNIT,player.cmd.sidemove * -FRACUNIT)
	
	if ctrlStyle(player) == 3
		abilityaim = mo.angle
	end
	
	//Depending on the sharpness of your turn, you will lose speed
	local anglediff = AngleFixed(abs(momangle - abilityaim)) / FRACUNIT
	local speedpercentage = max(0, min(100, (100 - (anglediff - 50))))
	speed = $ * speedpercentage / 100
	
	
	//print(AngleFixed(momangle) / FRACUNIT)
	//print(AngleFixed(abilityaim) / FRACUNIT)
	//print(AngleFixed(mo.angle) / FRACUNIT)
	//print(anglediff)
	
	//Flags, states
	mo.dropcharge = 0
	mo.djready = false
	player.pflags = $ | PF_JUMPED
	player.pflags = $ | PF_STARTJUMP
	mo.state = S_PLAY_ROLL
	
	//Underwater and super speed modifiers
	if mo.eflags&MFE_UNDERWATER
		djspeed = $*2/3
		djmomz = $*2/3
	end
	if player.powers[pw_super]
		djspeed = $*3/2
		djmomz = $*3/2
		local dustamt = 6
	end
	
	//Apply vertical thrust
	P_SetObjectMomZ(mo, djmomz, false)
	
	//Apply horizontal thrust unless the movement controls are mostly neutral
	if (player.freeaim and FixedHypot(abs(player.cmd.forwardmove), abs(player.cmd.sidemove)) > 5)
	or (not player.freeaim and player.cmd.forwardmove > 5)
	or mo.dropcharge >= DROP_LV2TIME
		if not player.freeaim //Always face forwards
			player.drawangle = mo.angle
			abilityaim = mo.angle
			P_InstaThrust(mo, mo.angle, max(djspeed,speed))
			
		else//Free aim
			player.drawangle = abilityaim
			P_InstaThrust(mo, abilityaim, max(djspeed,speed))
		end
	end
	
	//Dust
	local function r(mo,value)
		local w = mo.radius/FRACUNIT/2
		return value + P_RandomRange(-w,w) * FRACUNIT
	end
	
	for n = 1,dustamt
		local d = P_SpawnMobj(r(mo,mo.x),r(mo,mo.y),mo.z,MT_SPINDUST)
		d.extravalue1 = FRACUNIT
		d.extravalue2 = FRACUNIT
		if P_MobjFlip(mo) == -1 d.eflags = $|MFE_VERTICALFLIP end
		P_SetObjectMomZ(d,FRACUNIT*2)
		local l = 180*n/dustamt
		P_InstaThrust(d,abilityaim+ANGLE_135+l*ANG1,DJ_SPEED/2)
	end
	return
end

//Tails flight activated
local function tailsFly(player)
	local mo = player.mo
	local speed = FixedHypot(mo.momx, mo.momy)
	local momangle = R_PointToAngle2(0,0,mo.momx,mo.momy)
	
	//print(mo.flyamount)
	if mo.flyamount <= 0 return true end
	
	//Momentum
	mo.momz = $ / 2
	P_Thrust(mo, momangle + ANGLE_180, speed / 4)
	
	//Flags, states
	mo.flytimer = 0
	mo.flyamount = $ - 1
	player.pflags = $ & ~PF_JUMPED
	player.pflags = $ & ~PF_SPINNING
	player.pflags = $ | PF_THOKKED
	mo.state = S_PLAY_FLY
	return true
end

//Air abilities
local function useAirAbility(player)
	if player.mo.skin == "sonic" sonicDoubleJump(player) end
	if player.mo.skin == "tails" tailsFly(player) end
	return false
end

//Various controls for all characters
addHook("ThinkFrame", function(player)
	for player in players.iterate()
		//Init
		if player.init == nil
			player.init = true
			player.prevbuttons = player.cmd.buttons
			if player.freeaim == nil player.freeaim = 1 end
			if player.uncurltoggle == nil player.uncurltoggle = 1 end
			resetAbilities(player)
			return
		end
		
		//print(ctrlStyle(player))
		
		//Buffed spindash
		if player.mindash < 20 * FRACUNIT
			player.mindash = 20 * FRACUNIT
		end
		
		//Lots of moveset tweaks
		if player.mo and player.playerstate == PST_LIVE and not player.exiting and not player.powers[pw_nocontrol] and not P_PlayerInPain(player)
			local pbtn = player.prevbuttons
			local btn = player.cmd.buttons
			local mo = player.mo
			local flip = P_MobjFlip(mo)
			local nocontrol = (P_PlayerInPain(player) or player.state == S_PLAY_PAIN or (player.playerstate == PST_DEAD) or player.powers[pw_carry])
			
			if (not mo.init)
				mo.init = true
				resetAbilities(player)
			end
			
			if P_IsObjectOnGround(mo) or (player.mo.eflags & MFE_SPRUNG)or (player.mo.eflags & MFE_JUSTSTEPPEDDOWN) or nocontrol
				resetAbilities(player)
			end
			
			//Uncurl
			if mo.state == S_PLAY_ROLL and P_IsObjectOnGround(mo) and not nocontrol
				local uncurlinput = false
				local recurlinput = false
				if player.uncurltoggle == 1
					uncurlinput = (btn & BT_USE) and not (pbtn & BT_USE)
				else
					uncurlinput = (pbtn & BT_USE) and not (btn & BT_USE)
					recurlinput = (btn & BT_USE) and not (pbtn & BT_USE)
				end
				
				if mo.uncurl_lock == UNCURL_LOCKTIME and uncurlinput
					uncurlinput = false
				end
				
				if uncurlinput
					mo.will_uncurl = true
				end
				if recurlinput
					mo.will_uncurl = false
				end
				
				if mo.uncurl_lock
					mo.uncurl_lock = max(0, $ - 1)
					
				elseif mo.will_uncurl
					mo.uncurl_lock = UNCURL_LOCKTIME
					mo.uncurlready = false
					if FixedHypot(mo.momx, mo.momy) >= player.runspeed
						mo.state = S_PLAY_RUN
					else
						mo.state = S_PLAY_WALK
					end
					if player.pflags & PF_SPINNING
						player.pflags = $1 & ~PF_SPINNING
					end
				end
			else
				mo.uncurl_lock = UNCURL_LOCKTIME
				mo.will_uncurl = false
			end
			
			//Full air control when jumping out of a roll for all characters
			if player.pflags & PF_SPINNING and not P_IsObjectOnGround(mo)
				if player.pflags & PF_JUMPED
					player.pflags = $ & ~PF_SPINNING
				end
			end
			
			//New air actions for all characters
			if not P_IsObjectOnGround(mo) and mo.state != S_PLAY_JUMP and not (player.pflags & PF_THOKKED) and not nocontrol
				//Double jump to save yourself
				if (btn & BT_JUMP) and not (pbtn & BT_JUMP)
					//Custom abilities
					if player.charability == CA_NONE
						useAirAbility(player)
						
					//Glide
					elseif player.charability == CA_GLIDEANDCLIMB
						local glidespeed = player.actionspd

						player.pflags = $ | PF_GLIDING
						player.glidetime = 0

						if player.powers[pw_super]
							// Glide at double speed while super.
							glidespeed = $ * 2
						else
							player.pflags = $ | PF_THOKKED
						end
						
						mo.state = S_PLAY_GLIDE
						P_InstaThrust(player.mo, player.mo.angle, FixedMul(glidespeed, player.mo.scale))
						player.pflags = $ & ~(PF_SPINNING|PF_STARTDASH)
					
					//Twinspin
					elseif player.charability == CA_TWINSPIN
						mo.state = S_PLAY_TWINSPIN
						player.pflags = $ | PF_THOKKED
					
					//Bounce
					elseif player.charability == CA_BOUNCE
						player.mo.state = S_PLAY_BOUNCE
						player.pflags = $ & ~(PF_JUMPED|PF_NOJUMPDAMAGE);
						player.pflags = $ | PF_THOKKED|PF_BOUNCING;
					
					//Float
					elseif player.charability == CA_FLOAT
						player.secondjump = 1
						mo.state = S_PLAY_STND
					end
					
				//Midair curl (unless can't spin)
				elseif (btn & BT_USE) and not (pbtn & BT_USE) and not (player.pflags & PF_SPINNING) and player.charability2 == CA2_SPINDASH and mo.state != S_PLAY_ROLL
					if mo.state == S_PLAY_FLY and mo.skin == "tails"
						mo.flyamount = 0
					end
					
					mo.state = S_PLAY_JUMP
					player.pflags = $ | PF_JUMPED
					player.pflags = $ & ~PF_THOKKED
					S_StartSound(mo, sfx_curl)
					
				//Twinspin using spin button
				elseif (btn & BT_USE) and not (pbtn & BT_USE) and not (player.pflags & PF_THOKKED) and player.charability == CA_TWINSPIN
					//Twinspin
					mo.state = S_PLAY_TWINSPIN
					player.pflags = $ | PF_THOKKED
				end
				
			end
			
			//Sonic's drop dash
			if mo.skin == "sonic"
				if player.charability == CA_THOK
					player.charability = CA_NONE
				end
				
				//Release drop dash
				if mo.eflags & MFE_JUSTHITFLOOR and mo.dropcharge >= DROP_LV1TIME and (btn & BT_USE)
				and (player.powers[pw_shield] == SH_NONE or player.powers[pw_shield] == SH_PITY)
				and not (player.mo.eflags & MFE_SPRUNG)
					S_StartSound(mo, sfx_zoom)
					
					local speed = FixedHypot(mo.momx, mo.momy)
					local dropspeed = DROP_LV1SPEED
					local dustamt = 4
					local abilityaim = mo.angle + R_PointToAngle2(0,0,player.cmd.forwardmove * FRACUNIT,player.cmd.sidemove * -FRACUNIT)
					
					if ctrlStyle(player) == 3
						abilityaim = mo.angle
					end
					
					if (player.freeaim and FixedHypot(abs(player.cmd.forwardmove), abs(player.cmd.sidemove)) <= 5)
						abilityaim = player.drawangle
					end
					
					//Charge level boosts, sfx, etc.
					if mo.dropcharge >= DROP_LV2TIME and mo.dropcharge < DROP_LV3TIME
						dropspeed = DROP_LV2SPEED
						S_StartSound(mo, sfx_s3k81)
					elseif mo.dropcharge >= DROP_LV3TIME
						dropspeed = DROP_LV3SPEED
						S_StartSound(mo, sfx_s3k81)
						S_StartSound(mo, sfx_s3ka2)
					end
					
					if not player.freeaim player.drawangle = mo.angle end//Always face forwards when charging dropdash
					
					//State change
					mo.state = S_PLAY_ROLL
					player.pflags = $1|PF_SPINNING
					
					//Thrust
					if mo.dropcharge >= DROP_LV2TIME dropspeed = DROP_LV2SPEED
					elseif mo.dropcharge >= DROP_LV3TIME dropspeed = DROP_LV3SPEED end
					if mo.eflags&MFE_UNDERWATER dropspeed = $*2/3 end
					if player.powers[pw_super] dropspeed = $*3/2 end
					
					if not player.freeaim //Always face forwards
						P_InstaThrust(mo, mo.angle, max(dropspeed,speed))
						
					else//Free aim
						P_InstaThrust(mo, abilityaim, max(dropspeed,speed))
					end
					
					//Visual, SFX
					
					//Dust
					if mo.dropcharge >= DROP_LV2TIME dustamt = 6
					elseif mo.dropcharge >= DROP_LV3TIME dustamt = 8 end
					
					local function r(mo,value)
						local w = mo.radius/FRACUNIT/2
						return value + P_RandomRange(-w,w) * FRACUNIT
					end
					
					for n = 1,dustamt
						local d = P_SpawnMobj(r(mo,mo.x),r(mo,mo.y),mo.z,MT_SPINDUST)
						d.extravalue1 = FRACUNIT
						d.extravalue2 = FRACUNIT
						if P_MobjFlip(mo) == -1 d.eflags = $|MFE_VERTICALFLIP end
						P_SetObjectMomZ(d,FRACUNIT*2)
						local l = 90*n/dustamt
						P_InstaThrust(d,player.drawangle+ANGLE_135+l*ANG1,dropspeed/2)
					end
					
					//Quake
					if mo.dropcharge >= DROP_LV3TIME P_StartQuake(30*FRACUNIT, 5, {mo.x,mo.y,mo.z}, 200 * FRACUNIT) end// [mo.x,mo.y,mo.z, 120 * FRACUNIT])
					
				//Charge drop dash
				elseif player.pflags & PF_JUMPED and btn & BT_USE and (not (pbtn & BT_USE) or mo.dropcharge > 0)
				and (player.powers[pw_shield] == SH_NONE or player.powers[pw_shield] == SH_PITY) and not nocontrol
					mo.dropcharge = min($ + 1, DROP_LV3TIME + 1)
					
					//Drop dash is ready
					if mo.dropcharge == DROP_LV1TIME
						S_StartSound(mo, sfx_drpdsh)
						mo.state = S_DROPDASH1
					end
					
					//Super drop dash is ready
					if mo.dropcharge >= DROP_LV2TIME and mo.dropcharge < DROP_LV3TIME
						if mo.dropcharge == DROP_LV2TIME S_StartSound(mo, sfx_s3kab8) end
						local trail = P_SpawnGhostMobj(mo)
						P_TeleportMove(trail, trail.x, trail.y, trail.z - 3 * FRACUNIT)
						trail.fuse = 8
						trail.state = S_THOK
						trail.scale = FRACUNIT * 4/5 
						trail.destscale = FRACUNIT/4
						trail.frame = TR_TRANS80|A
					
					//Ultra drop dash is ready
					elseif mo.dropcharge >= DROP_LV3TIME
						if mo.dropcharge == DROP_LV3TIME S_StartSound(mo, sfx_s3kabf) end
						local trail = P_SpawnGhostMobj(mo)
						P_TeleportMove(trail, trail.x, trail.y, trail.z - 7 * FRACUNIT)
						trail.fuse = 12
						trail.state = S_THOK
						trail.scale = FRACUNIT
						trail.destscale = FRACUNIT/4
						//trail.color = SKINCOLOR_CYAN
						//trail.colorized = true
					end
					if not player.freeaim player.drawangle = mo.angle end//Always face forwards when charging dropdash
				
				//Not charging anymore - exit rev animation
				else
					if mo.dropcharge >= DROP_LV1TIME and ((player.pflags & PF_SPINNING) or (player.pflags & PF_JUMPED))
					then mo.state = S_PLAY_ROLL end
					mo.dropcharge = 0
				end
			end
			
			//Tails hover physics
			if mo.skin == "tails"
				if player.charability == CA_FLY
					player.charability = CA_NONE
				end
				if mo.eflags&MFE_UNDERWATER
					mo.flyamount = FLY_AMOUNT
				end
				
				if mo.state == S_PLAY_FLY or mo.state == S_PLAY_FLY_TIRED or mo.state == S_PLAY_SWIM
					local flymomz = FLY_MOMZ
					local flyhover = FLY_HOVER
					local momztics = FLY_MOMZTICS
					if mo.eflags & MFE_UNDERWATER
						momztics = FLY_MOMZTICS_SWIM
					end
					local speed = FixedHypot(mo.momx, mo.momy)
					
					//print(speed/FRACUNIT)
					
					if mo.flytimer < FLY_TIREDTIME
						if mo.flytimer % 10 == 1 and mo.state == S_PLAY_FLY
							S_StartSound(mo, sfx_s3kba)
						end
			
					elseif mo.flytimer >= FLY_TIREDTIME
						if mo.flytimer % 5 == 1 and mo.state == S_PLAY_FLY
							S_StartSound(mo, sfx_s3kbb)
						end
						flyhover = FLY_HOVER_TIRED
					end
					
					if (flip == 1 and mo.momz < flyhover) or (flip == -1 and mo.momz > -flyhover)
						P_SetObjectMomZ(mo, flyhover, false)
					end
					//P_SetObjectMomZ(mo, min(flippedmomz, flyhover))
					
					
					mo.flytimer = $ + 1
					if (mo.flytimer <= momztics)
						//Momentum
						if mo.eflags&MFE_UNDERWATER
							flymomz = $*2/3
						end
						P_SetObjectMomZ(mo, flymomz, true)//apply vertical boost
						player.pflags = $ | PF_THOKKED
					
					elseif mo.flytimer >= FLY_MAXTIME or ((mo.eflags&MFE_UNDERWATER) and mo.flytimer >= FLY_MAXTIMEWATER)
						mo.flytimer = 0
						mo.state = S_PLAY_FALL
						
					elseif mo.flytimer >= FLY_SPAMTICS
						player.pflags = $ & ~PF_THOKKED
					end
				end
			end
			
			//Knuckles momentum renewal
			if player.charability == CA_GLIDEANDCLIMB then
				//Create glide history
				if player.glidelast == nil then
					player.glidelast = 0
				end
				local pmomx = player.rmomx
				local pmomy = player.rmomy //Current speed, scaled for normalspeed calculations
				local gliding = player.pflags&PF_GLIDING
				local thokked = player.pflags&PF_THOKKED
				local exitglide = (player.glidelast == 1 and not(gliding) and thokked)
				local landglide = (player.glidelast == 2 and not(gliding|thokked))
				//Restore glide momentum after deactivation
				if landglide then
					mo.momx = pmomx
					mo.momy = pmomy
				end
				//Update glide history
				if gliding then
					player.glidelast = 1 //Gliding
				elseif exitglide then
					player.glidelast = 2 //Falling from glide
				elseif not(gliding|thokked) then
					player.glidelast = 0 //Not in glide state
				end
			end
			
			player.prevbuttons = btn
			if player.powers[pw_carry]
				player.prevbuttons = $ | BT_JUMP
			end
		end
	end
end)

//Regain abilities after hitting enemy/boss
addHook("TouchSpecial", function(touched, mo)
	if not (mo.player and mo.player.valid) return end
	
	local player = mo.player
	if ((touched.flags & (MF_ENEMY|MF_BOSS))
	and (touched.flags & MF_SHOOTABLE) and not (touched.flags & MF_MISSILE))
	and (P_PlayerCanDamage(player, touched))
		if (touched.flags & MF_BOSS) or (touched.health > 1)//Reduce x and y velocity when bouncing off a boss / facestabber / etc
			mo.momx = $ * BOSS_MOMENTUM_PERCENT / 100
			mo.momy = $ * BOSS_MOMENTUM_PERCENT / 100
		end
		resetAbilities(player)
	end
end)

//New jump abilities
addHook("AbilitySpecial", function(player)
	if not (player.pflags & PF_JUMPED) or not (player.mo) return end
	local mo = player.mo
	
	//New air abilities
	if player.charability == CA_NONE and player.skin != "tails"
		if player.pflags & PF_THOKKED
			return false
		end
		useAirAbility(player)
		return true
	end
	
	//Fang momentum
	if player.charability == CA_BOUNCE
		if player.pflags & PF_THOKKED
			return false
		end
		player.mo.state = S_PLAY_BOUNCE
		player.pflags = $ & ~(PF_JUMPED|PF_NOJUMPDAMAGE);
		player.pflags = $ | PF_THOKKED|PF_BOUNCING;
		return true
	end
end)